
/*
############################################################################
#    Copyright (C) 2008 by Andrianov Andrew                                #
#    aifiltr0@gmail.com                                                    #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 3 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include "macros.h"

static int version_check(xmlNode * node1, xmlDoc* doc1)
{
    xmlNode *cur_node = NULL;
    xmlChar *str;
    for (cur_node = node1; cur_node; cur_node = cur_node->next)
    {
        if (cur_node->type == XML_ELEMENT_NODE)
        {
          //  DBG("node type: Element, name: %s", cur_node->name);
            if (xmlStrcmp(cur_node->name,(const xmlChar* ) "kopete-history")==0)
            {
                str = xmlGetProp(cur_node, "version");
               // DBG("Kopete log version: %s", str);
                if (xmlStrcmp(str,(const xmlChar* ) "0.9")==0)
                {
             //       DBG("Version is supported");
                    xmlFree(str);
                    return 0;
                }
                else
                {
                    xmlFree(str);
                    ERR("Version mismatch.");
                    return 1;
                }

            }
        }
        //To understand recursion one must
        //   return version_check(cur_node->children, doc1);
    }
}


int validate(char* path)
{
	xmlDoc *doc1 = NULL;
	xmlDoc *doc2 = NULL;
	xmlIndentTreeOutput = 1;
	xmlKeepBlanksDefault(0) ;
	xmlNode *root_element1 = NULL;
	xmlNode *root_element2 = NULL;
	LIBXML_TEST_VERSION
	
	/*parse the file and get the DOM */
    //doc1 = xmlReadFile(log1, NULL, 0);
    //doc2 = xmlReadFile(log2, NULL, 0);
	xmlParserCtxt* ctxt1;
	xmlParserCtxt* ctxt2;/* the parser context */
	/* create a parser context */
	ctxt1 = xmlNewParserCtxt();
	if (ctxt1 == NULL)
	{
		ERR("Failed to allocate parser context");
		return 1;
	}
	/* parse the file, activating the DTD validation option */
	doc1 = xmlCtxtReadFile(ctxt1, path, NULL, 0);
	/* check if parsing suceeded */
	if (doc1 == NULL)
	{
		ERR("Error parsing file %s (Will still do the trick if the file is empty)",path);
		
	}
	root_element1 = xmlDocGetRootElement(doc1);
    
	if (version_check(root_element1, doc1)!=0)
	{
		ERR("File %s is INVALID!!!", path);
	}

}

static int is_msg_present(xmlChar *time, xmlChar *msg, xmlDoc* doc)
{
   // DBG("Looking up...");
    xmlNode *node = xmlDocGetRootElement(doc);
    xmlNode *cur_node = NULL;
    xmlChar *st;
    xmlChar *txt;
    for (cur_node = node->children; cur_node; cur_node = cur_node->next)
    {
        if (xmlStrcmp(cur_node->name,(const xmlChar* ) "msg")==0)
        {
            st = xmlGetProp(cur_node, "time");
            
            if (xmlStrcmp(time,st)==0)
            {
                xmlFree(st);
		txt = xmlNodeListGetString(doc, cur_node->xmlChildrenNode, 1);
		if (xmlStrcmp(txt,msg)==0)
		{
			return 1;
		}
		xmlFree(txt);
            }
            else
            {
                xmlFree(st);
		
            }

        }
        //DBG("node type: Element, name: %s", cur_node->name);

    }
   // DBG("Prop NOT present");
    return 0;
}



//This is really dumb code. It scans a date and converts it to a number of seconds that is easy to compare
unsigned long long d2s(char* text)
{
    unsigned long long day,hour,min,sec;
    sscanf(text, "%llu %llu:%llu:%llu", &day,&hour,&min,&sec);
    return (day*24*60*60)+(hour*60*60)+(min*60)+sec;

}

//Returns a node that will hold the next log message
xmlNode * seeknsert(xmlNode * start_node, xmlDoc* doc, xmlNode* last_insert_node, xmlChar* time_m)
{
    xmlChar* time;
    xmlNode * node;
    xmlNode *cur_node=NULL;
    xmlNode *lstmsg=NULL;
    unsigned long long time_mi = d2s(time_m);
    if (last_insert_node!=NULL)
        node = last_insert_node;
    else
        node = start_node;

    node = start_node;
    for (cur_node = node->children; cur_node; cur_node = cur_node->next)
    {
        if (cur_node->type == XML_ELEMENT_NODE)
        {
            lstmsg=cur_node;
            //DBG("node type: Element, name: %s", cur_node->name);
            if (xmlStrcmp(cur_node->name,(const xmlChar* ) "msg")==0)
            {
                time = xmlGetProp(cur_node, "time");
                if (d2s(time)>time_mi)
                {
                    //DBG("Found a nice place =)");

                    return cur_node;
                }
            }
        }

	
    }
   // DBG("Loops finished, seems we're at the end");
    return NULL;
}

//This code looks kind-a weird.
//Proceed with caution or you may damage your brain

static int merge_trees(xmlNode * node1, xmlDoc* doc1, xmlNode * node2, xmlDoc* doc2)
{

    int cnt=0;;
    xmlNode *cur_node = NULL;
    xmlNode *buf;
    xmlNode * li = NULL;
    //xmlChar* str;

    xmlChar* time;
    xmlChar* text;
    xmlChar* in;
    xmlChar* from;
    xmlChar* nick;
    xmlNode *	next;
    xmlChar* txtbuf;
    for (cur_node = node1->children; cur_node; cur_node = cur_node->next)
    {
        if (cur_node->type == XML_ELEMENT_NODE)
        {
          //  DBG("node type: Element, name: %s", cur_node->name);
            if (xmlStrcmp(cur_node->name,(const xmlChar* ) "msg")==0)
            {
                //DBG("This seems to be the message record... Do we have that in the other file?");
                text = xmlNodeListGetString(doc1, cur_node->xmlChildrenNode, 1);
                time = xmlGetProp(cur_node, "time");
		in = xmlGetProp(cur_node, "in");
		from = xmlGetProp(cur_node, "from");
		nick = xmlGetProp(cur_node, "nick");
                //DBG("msg time: %s", time);
                if (is_msg_present(time, text, doc2))
                {
                  //  DBG("We do have it in the other tree!");
                }
                else
                {
                    //DBG("We do NOT have it in the other tree");
                    //merge here
                    //First let's find a place where to insert the element:
                    li = seeknsert(node2,doc2,li,time);
		   // DBG("Seek done;");
                    buf = xmlNewTextChild (node2, NULL, "msg", text);
                    //buf = xmlNewNode(NULL, (const xmlChar *) "msg");
                    //buf = xmlCopyNode(cur_node,2);
                    //INDIA-STYLE!!!!!
                    //  txtbuf=malloc();
                    xmlNewProp (buf, "nick", nick);
                    xmlNewProp (buf, "in", in);
                    xmlNewProp (buf, "from", from);
                    xmlNewProp (buf, "time", time);
		    if (li!=NULL)
		    xmlAddPrevSibling(li, buf);

		    /*next = li->next; 
		    li->next=buf;
		    buf->children=li->children;
		    buf->next=NULL;
		    //buf->last=NULL;
		    buf->next=li->next;
		    buf->parent=node2;
		    buf->prev=li;*/
                    //memcpy(buf, cur_node, sizeof(cur_node));
                    
                    cnt++;
                }
                xmlFree(time);
            }
        }
    }
    printf("Merge completed: %u elements merged\n", cnt);

}




// 1 update only log1 file, 2 - update only log2 file, 3 - update both (sync)
int merge(char* log1, char* log2, int update_method)
{

    DBG("Merging 2 files.");
    xmlDoc *doc1 = NULL;
    xmlDoc *doc2 = NULL;
    xmlIndentTreeOutput = 1;
    xmlKeepBlanksDefault(0) ;
    xmlNode *root_element1 = NULL;
    xmlNode *root_element2 = NULL;
    LIBXML_TEST_VERSION
    DBG("parsing...");
    /*parse the file and get the DOM */
    //doc1 = xmlReadFile(log1, NULL, 0);
    //doc2 = xmlReadFile(log2, NULL, 0);
    xmlParserCtxt* ctxt1;
    xmlParserCtxt* ctxt2;/* the parser context */
    /* create a parser context */
    ctxt1 = xmlNewParserCtxt();
    ctxt2 = xmlNewParserCtxt();
    if (ctxt1 == NULL)
    {
        ERR("Failed to allocate parser context");
        return 1;
    }
    if (ctxt2 == NULL)
    {
        ERR("Failed to allocate parser context");
        return 1;
    }
    /* parse the file, activating the DTD validation option */
    doc1 = xmlCtxtReadFile(ctxt1, log1, NULL, 0);
    /* check if parsing suceeded */
    if (doc1 == NULL)
    {
        ERR("Error parsing file %s (Will still do the trick if the file is empty)",log1);
        if ((ctxt1->lastError.domain==XML_FROM_PARSER) & (ctxt1->lastError.code==XML_ERR_DOCUMENT_EMPTY))
        {
            DBG("File is empty, the other tree will be copied");
            doc1 = xmlCtxtReadFile(ctxt1, log2, NULL, 0);
        }
    }

    doc2 = xmlCtxtReadFile(ctxt2, log2, NULL, 0);
    /* check if parsing suceeded */
    if (doc2 == NULL)
    {
        ERR("Error parsing file %s (Will still do the trick if the file is empty)",log2);
        if ((ctxt2->lastError.domain==XML_FROM_PARSER) & (ctxt2->lastError.code==XML_ERR_DOCUMENT_EMPTY))
        {
            DBG("File is empty, the other tree will be copied");
            doc2 = xmlCtxtReadFile(ctxt2, log1, NULL, 0);
        }
    }

    /* free up the parser context */
    //xmlFreeParserCtxt(ctxt);

    if (doc1 == NULL)
    {
        ERR("error: could not parse file %s", log1);
        return 2;
    }
    if (doc2 == NULL)
    {
        ERR("error: could not parse file %s\n", log2);
        return 2;
    }
    DBG("Files parsed");
    /*Get the root element node */
    root_element1 = xmlDocGetRootElement(doc1);
    root_element2 = xmlDocGetRootElement(doc2);
    //FILE* tmp1 = fopen("/tmp/1.xml","w+");
    //FILE* tmp2 = fopen("/tmp/2.xml","w+");
    if ((version_check(root_element1, doc1)==0) && (version_check(root_element2, doc2)==0))
    {
        DBG("Checks passed, merging");
        switch (update_method)
        {
        case 1:
            DBG("Merging changes to the first file");
            merge_trees(root_element2, doc2, root_element1, doc1);
            xmlSaveFormatFileEnc(log1, doc1, "UTF-8", 1);
            //xmlDocDump(tmp1, doc1);
            break;
        case 2:
            DBG("Merging changes to the second file");
            merge_trees(root_element1, doc1, root_element2, doc2);
            xmlSaveFormatFileEnc(log2, doc2, "UTF-8", 1);
            //xmlDocDump(tmp2, doc2);
            break;
        case 3:
            DBG("syncing files");
            merge_trees(root_element2, doc2, root_element1, doc1);
            merge_trees(root_element1, doc1, root_element2, doc2);
            xmlSaveFormatFileEnc(log1, doc1, "UTF-8", 1);
            xmlSaveFormatFileEnc(log2, doc2, "UTF-8", 1);
            //xmlDocDump(tmp2, doc1);
            //xmlDocDump(tmp2, doc2);
            break;

        default :
            ERR("invalid update mode");
            return 1;
            break;

        }
        DBG("Done");
        //fclose(tmp1);
        //fclose(tmp2);
        //xmlDocDump(stdout, doc1);
        //xmlDocDump(stdout, doc2);
    }
    else
    {
        ERR("Once or more files are of incompatible version");
    }

    //xmlSaveToFilename	("/tmp/test.xml", "UTF-8", 1)
    /*free the document */
    xmlFreeDoc(doc1);
    xmlFreeDoc(doc2);

    /*
    *Free the global variables that may
    *have been allocated by the parser.
    */
    xmlCleanupParser();


    return 0;
}

